#include "ocapi.h"
#include "lauxlib.h"
#include <inttypes.h>
#include <stdbool.h>
#include <stdlib.h>

static inline uint32_t get_lowercase(uint32_t cp) {
  uint32_t ret;
  /* load Character Register, then retrieve lowercase version */
  asm("MCR p4, 0, %1, cr0, cr0\n\t"
      "MRC p4, 1, %0, cr0, cr0"
      :"=r"(ret)
      :"r"(cp));
  return ret;
}

static inline uint32_t get_uppercase(uint32_t cp) {
  uint32_t ret;
  /* load Character Register, then retrieve uppercase version */
  asm("MCR p4, 0, %1, cr0, cr0\n\t"
      "MRC p4, 2, %0, cr0, cr0"
      :"=r"(ret)
      :"r"(cp));
  return ret;
}

static inline int get_display_width(uint32_t cp) {
  int ret;
  /* load Character Register, then retrieve display width */
  asm("MCR p4, 0, %1, cr0, cr0\n\t"
      "MRC p4, 3, %0, cr0, cr0"
      :"=r"(ret)
      :"r"(cp));
  return ret;
}

#define ISCONT(x) (((x)&0xC0)==0x80)
/* returns number of bytes to skip */
static inline size_t utf8_to_32(const uint8_t* in, uint32_t* out, size_t rem) {
  size_t skip;
  bool valid = false;
  if(in[0] < 0x80) {
    skip = 1;
    *out = in[0];
    valid = true;
  }
  else if(in[0] < 0xC0) {
    /* invalid, fall through */
    skip =1;
  }
  else if(in[0] < 0xE0) {
    if(rem >= 2 && (in[1] & 0xC0) == 0x80) {
      skip = 2;
      *out = ((in[0]&0x1F)<<6) | (in[1]&0x3F);
      if(*out >= 1<<7) valid = true;
    } else skip = 1;
  }
  else if(in[0] < 0xF0) {
    if(rem >= 3 && ISCONT(in[1]) && ISCONT(in[2])) {
      skip = 3;
      *out = ((in[0]&0x0F)<<12) | ((in[1]&0x3F)<<6) | (in[2]&0x3F);
      // shortest encoding && not surrogate && not FFFE/FFFF
      if(*out >= 1<<11 && !(*out >= 0xD800 && *out <= 0xDFFF) && *out < 0xFFFE)
        valid = true;
    } else skip = 1;
  }
  else if(in[0] < 0xF8) {
    if(rem >= 4 && ISCONT(in[1]) && ISCONT(in[2]) && ISCONT(in[3])) {
      skip = 4;
      *out = ((in[0]&0x07)<<18) | ((in[1]&0x3F)<<12) | ((in[2]&0x3F)<<6)
        | (in[3]&0x3F);
      // shortest encoding && in Unicode range
      if(*out >= 1<<16 && *out <= 0x10FFFF) valid = true;
    } else skip = 1;
  } else skip = 1;
  /* do not decode 26- or 31-bit sequences as the only reason they'll be there
     is because someone is trying to be pointlessly clever, and the only
     advantage of handling them is getting fewer U+FFFDs */
  if(!valid) {
    /* invalid sequence */
    *out = 0xFFFD; // U+FFFD REPLACEMENT CHARACTER
  }
  return skip;
}

static inline bool invalid_codepoint(uint32_t codepoint) {
  return codepoint > 0x10FFFF || (codepoint >= 0xD800 && codepoint <= 0xDFFF)
    || codepoint == 0xFFFE || codepoint == 0xFFFF;
}

static inline size_t utf32_len_in_8(uint32_t codepoint) {
  if(invalid_codepoint(codepoint)) return utf32_len_in_8(0xFFFD);
  else if(codepoint < 1<<7) return 1;
  else if(codepoint < 1<<11) return 2;
  else if(codepoint < 1<<16) return 3;
  else return 4;
}

static size_t utf32_in_8(uint32_t codepoint, uint8_t* out) {
  if(invalid_codepoint(codepoint)) return utf32_in_8(0xFFFD, out);
  else if(codepoint < 1<<7) {
    out[0] = codepoint;
    return 1;
  }
  else if(codepoint < 1<<11) {
    out[0] = (codepoint>>6)|0xC0;
    out[1] = (codepoint&0x3F)|0x80;
    return 2;
  }
  else if(codepoint < 1<<16) {
    out[0] = (codepoint>>12)|0xE0;
    out[1] = ((codepoint>>6)&0x3F)|0x80;
    out[2] = (codepoint&0x3F)|0x80;
    return 3;
  }
  else {
    out[0] = (codepoint>>18)|0xF0;
    out[1] = ((codepoint>>12)&0x3F)|0x80;
    out[2] = ((codepoint>>6)&0x3F)|0x80;
    out[3] = (codepoint&0x3F)|0x80;
    return 4;
  }
}

static size_t char_count(const uint8_t* p, size_t rem) {
  size_t ret = 0;
  while(rem > 0) {
    uint32_t cp;
    size_t skip = utf8_to_32(p, &cp, rem);
    rem -= skip; p += skip;
    ++ret;
  }
  return ret;
}

static size_t lowercased_byte_count(const uint8_t* p, size_t rem) {
  size_t ret = 0;
  while(rem > 0) {
    uint32_t cp;
    size_t skip = utf8_to_32(p, &cp, rem);
    rem -= skip; p += skip;
    ret += utf32_len_in_8(get_lowercase(cp));
  }
  return ret;
}

static size_t uppercased_byte_count(const uint8_t* p, size_t rem) {
  size_t ret = 0;
  while(rem > 0) {
    uint32_t cp;
    size_t skip = utf8_to_32(p, &cp, rem);
    rem -= skip; p += skip;
    ret += utf32_len_in_8(get_uppercase(cp));
  }
  return ret;
}

static int unicode_lower(lua_State* L) {
  size_t length;
  const uint8_t* instring = (const uint8_t*)luaL_checklstring(L, 1, &length);
  /* size buffer for exact string length */
  size_t outlength = lowercased_byte_count(instring, length);
  uint8_t* outstring = malloc(outlength);
  if(!outstring) return luaL_error(L, "out of memory");
  /* copy characters from input to output, running them through CP4 first */
  size_t rem = length;
  const uint8_t* inp = instring;
  uint8_t* outp = outstring;
  while(rem > 0) {
    uint32_t cp;
    size_t skip = utf8_to_32(inp, &cp, rem);
    inp += skip; rem -= skip;
    cp = get_lowercase(cp);
    outp += utf32_in_8(cp, outp);
  }
  lua_pushlstring(L, (const char*)outstring, outlength);
  return 1;
}

static int unicode_upper(lua_State* L) {
  size_t length;
  const uint8_t* instring = (const uint8_t*)luaL_checklstring(L, 1, &length);
  /* size buffer for exact string length */
  size_t outlength = uppercased_byte_count(instring, length);
  uint8_t* outstring = malloc(outlength);
  if(outstring == NULL) return luaL_error(L, "out of memory");
  /* copy characters from input to output, running them through CP4 first */
  size_t rem = length;
  const uint8_t* inp = instring;
  uint8_t* outp = outstring;
  while(rem > 0) {
    uint32_t cp;
    size_t skip = utf8_to_32(inp, &cp, rem);
    inp += skip; rem -= skip;
    cp = get_uppercase(cp);
    outp += utf32_in_8(cp, outp);
  }
  lua_pushlstring(L, (const char*)outstring, outlength);
  free(outstring); // not needed anymore
  return 1;
}

static int unicode_char(lua_State* L) {
  int top = lua_gettop(L);
  size_t outlen = 0;
  for(int n = 1; n <= top; ++n) {
    lua_Integer codepoint = luaL_checkinteger(L, n);
    if(codepoint > 0x10FFFF || codepoint < 0)
      return luaL_error(L, "value out of range");
    outlen += utf32_len_in_8(codepoint);
  }
  uint8_t* outstring = malloc(outlen);
  if(outstring == NULL) return luaL_error(L, "out of memory");
  uint8_t* outp = outstring;
  for(int n = 1; n <= top; ++n) {
    lua_Integer codepoint = luaL_checkinteger(L, n);
    // assert(!(codepoint > 0x10FFFF || codepoint < 0))
    outp += utf32_in_8(codepoint, outp);
  }
  lua_pushlstring(L, (const char*)outstring, outlen);
  free(outstring);
  return 1;
}

static int unicode_len(lua_State* L) {
  size_t inlength;
  const uint8_t* instring = (const uint8_t*)luaL_checklstring(L, 1, &inlength);
  size_t ret = 0, rem = inlength;
  const uint8_t* p = instring;
  while(rem > 0) {
    uint32_t cp;
    size_t skip = utf8_to_32(p, &cp, rem);
    p += skip; rem -= skip;
    ++ret;
  }
  lua_pushinteger(L, ret);
  return 1;
}

/* len will never be > 4 */
static inline void tinycopy(uint8_t* out, const uint8_t* in, size_t len) {
  switch(len) {
  case 4:
    out[3] = in[3];
  case 3:
    out[2] = in[2];
  case 2:
    out[1] = in[1];
  case 1:
    out[0] = in[0];
  default:
    ;
  }
}

static int unicode_reverse(lua_State* L) {
  /* somewhat proud of my solution to this one */
  size_t inlength;
  const uint8_t* instring = (const uint8_t*)luaL_checklstring(L, 1, &inlength);
  size_t outlength = inlength; // always going to be true
  uint8_t* outstring = malloc(outlength);
  if(outstring == NULL) return luaL_error(L, "out of memory");
  size_t rem = inlength;
  const uint8_t* inp = instring;
  uint8_t* outp = outstring + outlength;
  while(rem > 0) {
    uint32_t cp;
    size_t skip = utf8_to_32(inp, &cp, rem);
    /* ignore decoded codepoint and copy the sequence directly */
    outp -= skip;
    tinycopy(outp, inp, skip);
    inp += skip; rem -= skip;
  }
  lua_pushlstring(L, (const char*)outstring, outlength);
  free(outstring);
  return 1;
}

static int unicode_sub(lua_State* L) {
  size_t inlength;
  const uint8_t* instring = (const uint8_t*)luaL_checklstring(L, 1, &inlength);
  lua_Integer start, end, len;
  start = luaL_checkinteger(L, 2);
  end = luaL_checkinteger(L, 3);
  len = char_count(instring, inlength);
  if(start < 0) start = len + start;
  else if(start > 0) start -= 1;
  if(end < 0) end = len + end + 1;
  if(start >= len) start = len-1;
  if(end > len) end = len;
  /* calculate the start and end positions */
  size_t rem = start;
  const uint8_t* inp = instring;
  while(rem > 0) {
    uint32_t cp;
    size_t skip = utf8_to_32(inp, &cp, rem);
    inp += skip; --rem;
    /* ignore codepoint */
  }
  rem = end-start;
  const uint8_t* startp = inp;
  while(rem > 0) {
    uint32_t cp;
    size_t skip = utf8_to_32(inp, &cp, rem);
    inp += skip; --rem;
    /* ignore codepoint */
  }
  const uint8_t* endp = inp;
  lua_pushlstring(L, (const char*)startp, endp-startp);
  return 1;
}

static int unicode_isWide(lua_State* L) {
  size_t inlength;
  const uint8_t* instring = (const uint8_t*)luaL_checklstring(L, 1, &inlength);
  if(inlength == 0) return luaL_error(L,"must provide at least one character");
  uint32_t cp;
  utf8_to_32(instring, &cp, inlength);
  lua_pushboolean(L, get_display_width(cp) > 1);
  return 1;
}

static int unicode_charWidth(lua_State* L) {
  size_t inlength;
  const uint8_t* instring = (const uint8_t*)luaL_checklstring(L, 1, &inlength);
  if(inlength == 0) return luaL_error(L,"must provide at least one character");
  uint32_t cp;
  utf8_to_32(instring, &cp, inlength);
  lua_pushinteger(L, get_display_width(cp));
  return 1;
}

static int unicode_wlen(lua_State* L) {
  size_t inlength;
  const uint8_t* instring = (const uint8_t*)luaL_checklstring(L, 1, &inlength);
  size_t ret = 0, rem = inlength;
  const uint8_t* p = instring;
  while(rem > 0) {
    uint32_t cp;
    size_t skip = utf8_to_32(p, &cp, rem);
    p += skip; rem -= skip;
    ret += get_display_width(cp);
  }
  lua_pushinteger(L, ret);
  return 1;
}

/* more error tolerant than the real one, maybe? */
static int unicode_wtrunc(lua_State* L) {
  size_t inlength;
  const uint8_t* instring = (const uint8_t*)luaL_checklstring(L, 1, &inlength);
  lua_Integer targetlen = luaL_checkinteger(L, 2), curlen = 0;
  size_t rem = inlength;
  const uint8_t* p = instring;
  while(rem > 0 && curlen < targetlen) {
    uint32_t cp;
    size_t skip = utf8_to_32(p, &cp, rem);
    p += skip; rem -= skip;
    curlen += get_display_width(cp);
  }
  lua_pushlstring(L, (const char*)instring, p-instring);
  return 1;
}

static const luaL_Reg unicodelib[] = {
  {"lower", unicode_lower},
  {"upper", unicode_upper},
  {"char", unicode_char},
  {"len", unicode_len},
  {"reverse", unicode_reverse},
  {"sub", unicode_sub},
  {"isWide", unicode_isWide},
  {"charWidth", unicode_charWidth},
  {"wlen", unicode_wlen},
  {"wtrunc", unicode_wtrunc},
  {NULL, NULL}
};

int ocapi_open_unicode(lua_State* L) {
  luaL_newlib(L, unicodelib);
  return 1;
}
