#include <stdlib.h>
#include <string.h>
#include "interchange.h"
#include "die.h"
#include "uuid.h"

#include <lauxlib.h>

/* currently always assumes INTERCHANGE_PACKED is not in use */

void* get_io_buffer(size_t size) {
  static void* io_buffer = NULL;
  static size_t io_buffer_size = 0;
  if(size > io_buffer_size) {
    void* ret = realloc(io_buffer, size);
    if(ret != NULL) {
      io_buffer = ret;
      io_buffer_size = size;
    }
    return ret;
  }
  return io_buffer;
}

int16_t ic_get_type(const void* p) {
  return *(int32_t*)p;
}

const void* ic_push(lua_State* L, const void* p) {
  int16_t type = ic_get_type(p);
  p += 4;
  switch(type) {
  case ICTAG_END:
    luaL_error(L, "cannot ic_push(ICTAG_END)!");
    __builtin_unreachable();
  case ICTAG_NULL: lua_pushnil(L); break;
  case ICTAG_BOOLEAN: lua_pushboolean(L, *(int32_t*)p != 0); p += 4; break;
  case ICTAG_DOUBLE: lua_pushnumber(L, *(double*)p); p += 8; break;
  case ICTAG_INT: lua_pushinteger(L, *(int32_t*)p); p += 4; break;
  case ICTAG_ARRAY: {
    lua_newtable(L);
    int i = 0;
    while((type = ic_get_type(p)) != ICTAG_END) {
      p = ic_push(L, p);
      lua_rawseti(L, ++i, -2);
    }
    p += 4;
  } break;
  case ICTAG_COMPOUND: {
    lua_newtable(L);
    while((type = ic_get_type(p)) != ICTAG_END) {
      p = ic_push(L, p);
      if((type = ic_get_type(p)) == ICTAG_END) {
        luaL_error(L, "malformed ICTAG_ARRAY");
        __builtin_unreachable();
      }
      p = ic_push(L, p);
      lua_settable(L, -3);
    }
    p += 4;
  } break;
  case ICTAG_UUID: lua_pushstring(L, uuid2staticstring(p)); p += 16; break;
  default:
    if(type < 0) {
      luaL_error(L, "unknown interchange tag");
      __builtin_unreachable();
    }
    /* we treat ICTAG_STRING and ICTAG_BYTE_ARRAY the same */
    int length = (type & 16383);
    void* ret = malloc(length);
    if(length > 0 && ret == NULL) {
      luaL_error(L, "out of memory");
      __builtin_unreachable();
    }
    memcpy(ret, p, length);
    if(length&3) length += 4-(length&3);
    p += length;
  }
  return p;
}

const void* ic_skip(const void* p) {
  int16_t type = ic_get_type(p);
  p += 4;
  switch(type) {
  case ICTAG_END:
    die("cannot ic_skip(ICTAG_END)!");
    __builtin_unreachable();
  case ICTAG_NULL: break;
  case ICTAG_BOOLEAN: p += 4; break;
  case ICTAG_DOUBLE: p += 8; break;
  case ICTAG_INT: p += 4; break;
  case ICTAG_ARRAY: {
    while((type = ic_get_type(p)) != ICTAG_END)
      p = ic_skip(p);
    p += 4;
  } break;
  case ICTAG_COMPOUND: {
    while((type = ic_get_type(p)) != ICTAG_END) {
      p = ic_skip(p);
      if((type = ic_get_type(p)) == ICTAG_END) {
        die("malformed ICTAG_ARRAY");
        __builtin_unreachable();
      }
      p = ic_skip(p);
    }
    p += 4;
  } break;
  case ICTAG_UUID: p += 16; break;
  default:
    if(type < 0) {
      die("unknown interchange tag");
      __builtin_unreachable();
    }
    /* we treat ICTAG_STRING and ICTAG_BYTE_ARRAY the same */
    int length = (type & 16383);
    if(length&3) length += 4-(length&3);
    p += length;
  }
  return p;
}
