#define DEBUG 0
#define HEAPCHECK 0

/* must be multiples of 4, ignored if HEAPCHECK is zero */
#define HEAPCHECK_HEADER 16
#define HEAPCHECK_FOOTER 16
#define HEAPCHECK_TOP_LEN 64
#define HEAPCHECK_WORD 0xCACACACA

#include <stdlib.h>
#include <inttypes.h>
#include <errno.h>
#include "die.h"

extern int __heap_start;
extern unsigned int ram_bytes;
#if HEAPCHECK
#define HEAP_START ((void*)(((uint8_t*)&__heap_start) + HEAPCHECK_TOP_LEN))
#define HEAPCHECK_START ((uint32_t*)&__heap_start)
#else
#define HEAP_START ((void*)&__heap_start)
#endif
#define HEAP_END ((void*)ram_bytes)

static struct blockheader {
  struct blockheader* prev_block;
  struct blockheader* next_block;
  size_t block_flags;
  uint8_t data[0];
  struct blockheader* prev_free;
  struct blockheader* next_free;
}* freelist;

#define BLOCK_MAIN_HEADER_BYTES 12
#define BLOCK_FREE_SUBHEADER_BYTES 8

#define GET_BLOCKHEADER(ptr) ((struct blockheader*)((uint8_t*)(ptr)-BLOCK_MAIN_HEADER_BYTES))
#define BLOCK_IS_FREE(block) ((block)->block_flags&1)
#define BLOCK_SET_FREE(block) ((block)->block_flags|=1)
#define BLOCK_SET_ALLOCATED(block) ((block)->block_flags&=~(size_t)1)
#define BLOCK_GET_SIZE(block) ((block)->block_flags&~(size_t)3)
// size must be a multiple of 4
#define BLOCK_SET_SIZE(block, new) ((block)->block_flags = ((block)->block_flags&3)|(new))

#if DEBUG || HEAPCHECK
#define GET_LINK_REGISTER() void* lr; asm volatile("MOV %0, lr":"=r"(lr):)
static void* real_malloc(size_t size);
static void* real_calloc(size_t a, size_t b);
static void* real_realloc(void* p, size_t size);
static void real_free(void* p);
#endif

size_t get_heap_free_bytes() {
  size_t ret = 0;
  for(struct blockheader* block = HEAP_START; block; block = block->next_block) {
    if(BLOCK_IS_FREE(block)) ret += BLOCK_GET_SIZE(block);
  }
  return ret;
}

size_t get_heap_size() {
  return HEAP_END - HEAP_START;
}

#if DEBUG
#include <stdbool.h>
#include <stdio.h>
static char buffer[160];
static void dputs(const char* buffer) {
  while(*buffer) {
    if(*buffer == '\n')
      asm volatile("CDP p7, 0, cr0, cr0, cr0, #0"::);
    else
      asm volatile("MCR p7, 0, %0, cr0, cr0"::"r"(*buffer));
    ++buffer;
  }
}
#define dprintf(...) do { sprintf(buffer,__VA_ARGS__);dputs(buffer); } while(0)
static void dump_block(struct blockheader* block) {
  dprintf("%p: prev(%p) next(%p) size(%#x,%u) type(%s)",
          block, block->prev_block, block->next_block,
          BLOCK_GET_SIZE(block),
          BLOCK_GET_SIZE(block), BLOCK_IS_FREE(block)?"free":"allocated");
  if(BLOCK_IS_FREE(block))
    dprintf(" prev_free(%p) next_free(%p)", block->prev_free,block->next_free);
#if HEAPCHECK
  else {
    bool valid = true;
    uint32_t* wptr = (uint32_t*)block->data;
    for(int n = 0; n < HEAPCHECK_HEADER; n += 4) {
      if(wptr[n/4] != HEAPCHECK_WORD) { valid = false; break; }
    }
    if(valid) {
      wptr = (uint32_t*)(block->data + BLOCK_GET_SIZE(block) - HEAPCHECK_FOOTER);
      for(int n = 0; n < HEAPCHECK_FOOTER; n += 4) {
        if(wptr[n/4] != HEAPCHECK_WORD) { valid = false; break; }
      }
    }
    if(!valid) {
      dprintf(" CORRUPTED\n");
      wptr = (uint32_t*)block->data;
      for(int n = 0; n < HEAPCHECK_HEADER; n += 4) {
        dprintf(" %08X", wptr[n/4]);
      }
      dprintf(" /");
      wptr = (uint32_t*)(block->data + BLOCK_GET_SIZE(block) - HEAPCHECK_FOOTER);
      for(int n = 0; n < HEAPCHECK_FOOTER; n += 4) {
        dprintf(" %08X", wptr[n/4]);
      }
    }
    else dprintf(" intact");
  }
#endif
  dputs("\n");

}
static void dump_heap() {
  dputs("--- HEAP DUMP ---\n");
#if HEAPCHECK
  uint32_t* p = HEAPCHECK_START;
  bool heap_corrupted = false;
  for(int n = 0; n < HEAPCHECK_TOP_LEN; n += 4) {
    if(p[n/4] != HEAPCHECK_WORD) {
      heap_corrupted = true;
      break;
    }
  }
  if(heap_corrupted) {
    dputs("HEAP TOP IS CORRUPTED\n");
    for(int n = 0; n < HEAPCHECK_TOP_LEN; n += 4) {
      dprintf(" %08X", p[n/4]);
    }
    dputs("\n");
  }
#endif
  dprintf("(heap extent: %p -- %p)\n", HEAP_START, HEAP_END);
  dprintf("freelist: %p\n", freelist);
  for(struct blockheader* block = HEAP_START; block; block = block->next_block) {
    dump_block(block);
  }
  dputs("--- END HEAP DUMP ---\n");
}
#else
#define dputs(buffer)
#define dprintf(...)
#define dump_block()
#define dump_heap()
#endif

#if HEAPCHECK
static void write_heapcheck(void* ptr) {
  struct blockheader* block = GET_BLOCKHEADER(ptr);
  uint32_t* wptr = (uint32_t*)ptr;
  for(int n = 0; n < HEAPCHECK_HEADER; n += 4) {
    wptr[n/4] = HEAPCHECK_WORD;
  }
  wptr = (uint32_t*)((uint8_t*)ptr + BLOCK_GET_SIZE(block) - HEAPCHECK_FOOTER);
  for(int n = 0; n < HEAPCHECK_FOOTER; n += 4) {
    wptr[n/4] = HEAPCHECK_WORD;
  }
  ptr = (uint8_t*)(ptr + HEAPCHECK_HEADER);
}
#endif

void init_heap(void) {
  if(ram_bytes <= (unsigned int)HEAP_START + sizeof(struct blockheader))
    die("no heap"); // THIS won't happen often
  freelist = (struct blockheader*)HEAP_START;
  /* make sure we are aligned to a 4-byte boundary */
  if((unsigned int)freelist&3)
    freelist = (struct blockheader*)((uint8_t*)freelist+(4-(unsigned int)freelist&3));
  freelist->prev_block = NULL;
  freelist->next_block = NULL;
  freelist->prev_free = NULL;
  freelist->next_free = NULL;
  BLOCK_SET_FREE(freelist);
  BLOCK_SET_SIZE(freelist, (ram_bytes - (unsigned int)freelist->data)&~(size_t)3);
  dputs("init_heap() called\n");
#if HEAPCHECK
  uint32_t* p = HEAPCHECK_START;
  for(int n = 0; n < HEAPCHECK_TOP_LEN; n += 4) {
    p[n/4] = HEAPCHECK_WORD;
  }
#endif
#if DEBUG
  dump_heap();
  dputs("\n");
#endif
}

/* first must come before second, the second ought to be free, and the caller's
   responsible for doing any twiddling of the free/block pointers */
static void merge_blocks(struct blockheader* first,
                         struct blockheader* second) {
  size_t temp = BLOCK_GET_SIZE(first) + BLOCK_GET_SIZE(second) + BLOCK_MAIN_HEADER_BYTES;
  BLOCK_SET_SIZE(first, temp);
}

static void remove_from_freelist(struct blockheader* block) {
  if(block->prev_free != NULL) block->prev_free->next_free = block->next_free;
  else freelist = block->next_free;
  if(block->next_free != NULL) block->next_free->prev_free = block->prev_free;
}

/* size is rounded up to the nearest multiple of 4 */
void* malloc(size_t size) {
#if DEBUG || HEAPCHECK
  GET_LINK_REGISTER();
#if DEBUG
  dprintf("%p: malloc(%#x,%u)\n", lr, size, size);
  dump_heap();
#endif
  void* ret;
#if HEAPCHECK
  if(size != 0) {
    if(size + (HEAPCHECK_HEADER + HEAPCHECK_FOOTER + 128) < size) {
      errno = ENOMEM;
      ret = NULL;
      goto skip_ret_being_set;
    }
    ret = real_malloc(size + HEAPCHECK_HEADER + HEAPCHECK_FOOTER);
  }
  else
#endif
    ret = real_malloc(size);
#if HEAPCHECK
  skip_ret_being_set:
  if(ret) write_heapcheck(ret);
#endif
#if DEBUG
  dprintf("malloc(%u) returned %p\n", size, ret);
  if(ret) dump_block(GET_BLOCKHEADER(ret));
  dump_heap();
  dputs("\n");
#endif
#if HEAPCHECK
  return ret ? ((uint8_t*)ret + HEAPCHECK_HEADER) : ret;
#else
  return ret;
#endif
}
#define malloc real_malloc
static void* real_malloc(size_t size) {
#endif
  if(size == 0) return NULL;
  if(size & 3) size = (size&~(size_t)3)+4;
  struct blockheader* block = freelist;
  while(block) {
    if(BLOCK_GET_SIZE(block) >= size + sizeof(struct blockheader)) {
      /* this block contains enough room for us AND enough room for another
         free block */
      struct blockheader* new_next = (struct blockheader*)block->data + size;
      /* we have been careful to lay out our structures so that no clobbering
         will take place here */
      remove_from_freelist(block);
      new_next->prev_free = NULL;
      new_next->next_free = freelist;
      freelist = new_next;
      BLOCK_SET_FREE(new_next);
      BLOCK_SET_SIZE(new_next, BLOCK_GET_SIZE(block) - sizeof(struct blockheader) - size);
      BLOCK_SET_SIZE(block, size);
      new_next->prev_block = (struct blockheader*)block;
      new_next->next_block = (struct blockheader*)block->next_block;
      block->next_block->prev_block = (struct blockheader*)new_next;
      block->next_block = (struct blockheader*)new_next;
      BLOCK_SET_ALLOCATED(block);
      return block->data;
    }
    else if(BLOCK_GET_SIZE(block) >= size) {
      /* this block contains enough room for us, butnot for another free
         block; we expand to fill the space */
      if(block->prev_free != NULL) block->prev_free->next_free = block->next_free;
      else freelist = block->next_free;
      if(block->next_free != NULL) block->next_free->prev_free = block->prev_free;
      BLOCK_SET_ALLOCATED(block);
      return block->data;
    }
    block = block->next_free;
  }
  errno = ENOMEM;
  return NULL;
}

void free(void* p) {
#if DEBUG || HEAPCHECK
  GET_LINK_REGISTER();
#if HEAPCHECK
  if(p) p = (uint8_t*)p - HEAPCHECK_HEADER;
#endif
#if DEBUG
  dprintf("%p: free(%p)\n", lr, p);
  if(p) dump_block(GET_BLOCKHEADER(p));
  dump_heap();
#endif
#if HEAPCHECK
  if(p != NULL)
    real_free((uint8_t*)p - HEAPCHECK_HEADER);
  else
#endif
    real_free(p);
#if DEBUG
  dump_heap();
  dputs("\n");
#endif
}
#define free real_free
static void real_free(void* p) {
#endif
  if(p == NULL) return;
  struct blockheader* block = GET_BLOCKHEADER(p);
  if(BLOCK_GET_SIZE(block) >= ram_bytes) die("heap corruption detected");
  if(BLOCK_IS_FREE(block)) die("double free");
  BLOCK_SET_FREE(block);
  if(block->next_block != NULL && BLOCK_IS_FREE(block->next_block)) {
    /* delete block->next_block from the freelist */
    remove_from_freelist(block->next_block);
    struct blockheader* next_next_block = block->next_block->next_block;
    if(next_next_block)
      next_next_block->prev_block = block;
    merge_blocks(block, block->next_block);
    block->next_block = next_next_block;
  }
  if(block->prev_block != NULL && BLOCK_IS_FREE(block->prev_block)) {
    /* the freelist is consistent if we just... disappear, merging into
       the previous block */
    block->prev_block->next_block = block->next_block;
    if(block->next_block != NULL)
      block->next_block->prev_block = block->prev_block;
    merge_blocks(block->prev_block, block);
  }
  else {
    /* we must explicitly add ourselves to the freelist */
    block->prev_free = NULL;
    block->next_free = freelist;
    freelist->prev_free = block;
    freelist = block;
  }
}

/*
  Zeroes memory.
  base is assumed to be word-aligned. size is (effectively) rounded up to the
  nearest multiple of four.
  Good thing this lines up with what malloc is allowed to return...
 */
extern void _wazero(void* base, size_t size);
/*
  Copies memory.
  Makes same assumptions as _wazero.
*/
extern void _wacopy(void* dst, const void* src, size_t size);

void* calloc(size_t a, size_t b) {
#if DEBUG || HEAPCHECK
  GET_LINK_REGISTER();
#if DEBUG
  dprintf("%p: calloc(%u, %u)\n", lr, a, b);
  dump_heap();
#endif
  void* ret = real_calloc(a, b);
#if DEBUG
  dump_heap();
  dprintf("calloc(%u, %u) returned %p\n", a, b, ret);
  if(ret) dump_block(GET_BLOCKHEADER(ret));
  dputs("\n");
#endif
  return ret;
}
static void* real_calloc(size_t a, size_t b) {
#endif
  unsigned long long full = (unsigned long long)a * (unsigned long long)b;
  if((full >> 31) != 0ULL) die("calloc overflow"); // rare, >2GB alloc
  void* ret = malloc((size_t)full);
  if(ret)
    _wazero(ret, (size_t)full);
  return ret;
}

void* realloc(void* ptr, size_t new_size) {
#if DEBUG || HEAPCHECK
  GET_LINK_REGISTER();
#if HEAPCHECK
  if(ptr) ptr = (uint8_t*)ptr - HEAPCHECK_HEADER;
#endif
#if DEBUG
  dprintf("%p: realloc(%p, %u)\n", lr, ptr, new_size);
  if(ptr) dump_block(GET_BLOCKHEADER(ptr));
  dump_heap();
#endif
  void* ret;
#if HEAPCHECK
  if(ptr != NULL) {
    if(new_size == 0)
      ret = real_realloc(ptr, 0);
    else
      ret = real_realloc(ptr, new_size + HEAPCHECK_HEADER + HEAPCHECK_FOOTER);
  }
  else if(new_size != 0)
    ret = real_realloc(ptr, new_size + HEAPCHECK_HEADER + HEAPCHECK_FOOTER);
  else
#endif
    ret = real_realloc(ptr, new_size);
#if HEAPCHECK
  write_heapcheck(ret);
#endif
#if DEBUG
  dump_heap();
  dprintf("realloc(%p, %u) returned %p\n", ptr, new_size, ret);
  if(ret) dump_block(GET_BLOCKHEADER(ret));
  dputs("\n");
#endif
#if HEAPCHECK
  return ret ? ((uint8_t*)ret + HEAPCHECK_HEADER) : ret;
#else
  return ret;
#endif
}
static void* real_realloc(void* ptr, size_t new_size) {
#endif
  if(ptr == NULL) {
    if(new_size == 0) return NULL;
    else return malloc(new_size);
  }
  else if(new_size == 0) {
    if(ptr) free(ptr);
    return NULL;
  }
  /* pad size to four bytes */
  if(new_size&3) new_size = new_size + (4-(new_size&3));
  struct blockheader* block = GET_BLOCKHEADER(ptr);
  size_t old_size = BLOCK_GET_SIZE(block);
  if(BLOCK_IS_FREE(block) || old_size >= ram_bytes)
    die("heap corruption detected");
  if(old_size == new_size) return ptr; /* no change needed */
  else if(old_size < new_size) {
    size_t gain = new_size - old_size;
    /* We are growing. Is the next block a free block? */
    if(block->next_block != NULL && BLOCK_IS_FREE(block->next_block)) {
      /* Three cases now.
         1. Next block does not contain enough space to contain us.
         ---> Fall through.
         2. Next block has space, but not enough for another freeblock after.
         ---> Eat the next block and grow.
         3. Next block has space, and enough for another freeblock after.
         ---> Eat the next block and spit out a new freeblock.
         2 and 3 are handled by eating the entire freeblock and then calling
         ourselves recursively to possibly shrink and create a freeblock out
         of the remaining space. */
      if(BLOCK_GET_SIZE(block->next_block) + BLOCK_MAIN_HEADER_BYTES >= gain) {
        /* 2 or 3. Eat it! */
        remove_from_freelist(block->next_block);
        if(block->next_block->next_block)
          block->next_block->next_block->prev_block = block;
        merge_blocks(block, block->next_block);
        block->next_block = block->next_block->next_block;
        /* then call ourselves recursively */
        return realloc(ptr, new_size);
      }
      else {
        /* 1; fall through */
      }
    }
    /* Next block was not free, or did not have enough space. */
    /* Note that we make no attempt to merge with the previous block. */
    /* malloc and copy! */
    void* new_ptr = malloc(new_size);
    if(!new_ptr) return NULL; /* Uh oh... */
    _wacopy(new_ptr, ptr, old_size);
    free(ptr);
    return new_ptr;
  }
  else {
    size_t loss = old_size - new_size;
    if(block->next_block != NULL && BLOCK_IS_FREE(block->next_block)) {
      /* Shrink, and grow the next block along with us. */
      struct blockheader* next_old = block->next_block;
      struct blockheader* next_new = (struct blockheader*)((uint8_t*)next_old - loss);
      next_new->prev_block = next_old->prev_block; /* us! */
      next_new->next_block = next_old->next_block;
      next_new->block_flags = next_old->block_flags + loss; /*safe, I promise*/
      next_new->prev_free = next_old->prev_free;
      next_new->next_free = next_old->next_free;
      /* Fix up the list pointers */
      block->next_block = next_new;
      if(next_new->next_block) next_new->next_block->prev_block = next_new;
      if(next_new->next_free) next_new->next_free->prev_free = next_new;
      if(next_new->prev_free) next_new->prev_free->next_free = next_new;
      else freelist = next_new;
      /* And shrink. */
      BLOCK_SET_SIZE(block, new_size);
    }
    else if(loss >= sizeof(struct blockheader)) {
      /* We can shrink; there's room for a new free block at the end. */
      BLOCK_SET_SIZE(block, new_size);
      /* Make that free block. */
      struct blockheader* new_free = (struct blockheader*)(block->data+new_size);
      new_free->prev_block = block;
      new_free->next_block = block->next_block;
      if(block->next_block) block->next_block->prev_block = new_free;
      block->next_block = new_free;
      new_free->prev_free = NULL;
      new_free->next_free = freelist;
      freelist = new_free;
      BLOCK_SET_SIZE(new_free, loss - BLOCK_FREE_SUBHEADER_BYTES);
      BLOCK_SET_FREE(new_free);
    }
    else {
      /* We can't shrink, as there isn't enough room.
         Oh well, we tried. */
    }
    return ptr;
  }
}
